@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/wwwoffle.c 2.51 2002/05/26 10:50:42 amb Exp $
+  $Header: /home/amb/CVS/wwwoffle/src/wwwoffle.c,v 2.85 2009-03-13 19:30:39 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7c.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.9f.
   A user level program to interact with the server.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1996-2009 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -23,22 +23,23 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#include "version.h"
 #include "wwwoffle.h"
-#include "document.h"
+#include "io.h"
 #include "misc.h"
+#include "proto.h"
+#include "errors.h"
 #include "config.h"
 #include "sockets.h"
-#include "errors.h"
+#include "document.h"
+#include "version.h"
+
 
 #ifndef PATH_MAX
+/*+ The maximum pathname length in characters. +*/
 #define PATH_MAX 4096
 #endif
 
 
-static void usage(int verbose);
-
-
 /*+ The action to perform. +*/
 typedef enum _Action
 {
@@ -52,6 +53,10 @@ typedef enum _Action
 
  Config,                        /*+ Tell the server to re-read the configuration file. +*/
 
+ Dump,                          /*+ Tell the server to dump the configuration file. +*/
+
+ CycleLog,                      /*+ Tell the server to close and re-open the log file. +*/
+
  Purge,                         /*+ Tell the server to purge pages. +*/
 
  Status,                        /*+ Find out from the server the current status. +*/
@@ -68,18 +73,22 @@ typedef enum _Action
 Action;
 
 
-static void add_url_list(char **links);
+/* Local functions */
+
+static void usage(int verbose);
+
+static void add_url_list(URL **links);
 static void add_url_file(char *url_file);
 
+
+/* Local variables */
+
 /*+ The list of URLs or files. +*/
-static char **url_file_list=NULL;
+static /*@null@*/ char **url_file_list=NULL;
 
 /*+ The number of URLs or files. +*/
 static int n_url_file_list=0;
 
-/*+ A file descriptor for the spool directory; not used anywhere, but required for linking with spool.o. +*/
-int fSpoolDir=-1;
-
 
 /*++++++++++++++++++++++++++++++++++++++
   The main program.
@@ -90,7 +99,7 @@ int main(int argc, char** argv)
  int i;
  int recursive_mode=0,depth=0;
  int force=0;
- int stylesheets=0,images=0,frames=0,scripts=0,objects=0;
+ int stylesheets=0,images=0,frames=0,iframes=0,scripts=0,objects=0,nothing=0;
  char *config_file=NULL;
  int exitval=0;
 
@@ -98,7 +107,7 @@ int main(int argc, char** argv)
 
  char *env=NULL;
  char *host=NULL;
- int port=0,http_port,wwwoffle_port;
+ int port=0;
 
  /* Parse the command line options */
 
@@ -107,6 +116,8 @@ int main(int argc, char** argv)
 
  for(i=1;i<argc;i++)
    {
+    /* Main controlling options */
+
     if(!strcmp(argv[i],"-h") || !strcmp(argv[i],"--help"))
        usage(1);
 
@@ -149,131 +160,6 @@ int main(int argc, char** argv)
        continue;
       }
 
-    if(!strcmp(argv[i],"-F"))
-      {
-       if(action!=None && action!=Get)
-         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
-       action=Get;
-       if(force)
-         {fprintf(stderr,"wwwoffle: Only one '-F' argument may be given.\n"); exit(1);}
-       force=1;
-       argv[i]=NULL;
-       continue;
-      }
-
-    if(!strncmp(argv[i],"-g",2))
-      {
-       if(action!=None && action!=Get)
-         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
-       action=Get;
-       if(strchr(argv[i]+2,'S'))
-         {
-          if(stylesheets)
-            {fprintf(stderr,"wwwoffle: Only one '-gS' argument may be given.\n"); exit(1);}
-          stylesheets=1;
-         }
-       if(strchr(argv[i]+2,'i'))
-         {
-          if(images)
-            {fprintf(stderr,"wwwoffle: Only one '-gi' argument may be given.\n"); exit(1);}
-          images=1;
-         }
-       if(strchr(argv[i]+2,'f'))
-         {
-          if(frames)
-            {fprintf(stderr,"wwwoffle: Only one '-gf' argument may be given.\n"); exit(1);}
-          frames=1;
-         }
-       if(strchr(argv[i]+2,'s'))
-         {
-          if(scripts)
-            {fprintf(stderr,"wwwoffle: Only one '-gs' argument may be given.\n"); exit(1);}
-          scripts=1;
-         }
-       if(strchr(argv[i]+2,'o'))
-         {
-          if(objects)
-            {fprintf(stderr,"wwwoffle: Only one '-go' argument may be given.\n"); exit(1);}
-          objects=1;
-         }
-       if(argv[i][2]==0)
-          fprintf(stderr,"wwwoffle: The '-g' option does nothing on its own.\n");
-
-       argv[i]=NULL;
-       continue;
-      }
-
-    if(!strncmp(argv[i],"-R",2) || !strncmp(argv[i],"-r",2) || !strncmp(argv[i],"-d",2))
-      {
-       if(action!=None && action!=Get)
-         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
-       action=Get;
-
-       if(depth)
-         {fprintf(stderr,"wwwoffle: Only one '-d', '-r' or '-R' argument may be given.\n"); exit(1);}
-
-       if(argv[i][1]=='d')
-          recursive_mode=1;
-       else if(argv[i][1]=='r')
-          recursive_mode=2;
-       else /* argv[i][1]=='R' */
-          recursive_mode=3;
-
-       if(argv[i][2])
-          depth=atoi(&argv[i][2]);
-       else
-          depth=1;
-       if(depth<=0)
-         {fprintf(stderr,"wwwoffle: The '-%c' argument may only be followed by a positive integer.\n",argv[i][1]); exit(1);}
-
-       argv[i]=NULL;
-       continue;
-      }
-
-    if(!strcmp(argv[i],"-p"))
-      {
-       char *hoststr,*portstr;
-
-       if(++i>=argc)
-         {fprintf(stderr,"wwwoffle: The '-p' argument requires a hostname and optionally a port number.\n"); exit(1);}
-
-       if(config_file)
-         {fprintf(stderr,"wwwoffle: The '-p' and '-c' options cannot be used together.\n"); exit(1);}
-
-       SplitHostPort(argv[i],&hoststr,&portstr);
-
-       if(portstr)
-         {
-          port=atoi(portstr);
-
-          if(port<=0 || port>=65536)
-            {fprintf(stderr,"wwwoffle: The port number %d '%s' is invalid.\n",port,argv[i]); exit(1);}
-         }
-
-       host=hoststr;
-
-       /* Don't need to use RejoinHostPort(argv[i],hoststr,portstr) here. */
-
-       argv[i-1]=NULL;
-       argv[i]=NULL;
-       continue;
-      }
-
-    if(!strcmp(argv[i],"-c"))
-      {
-       if(++i>=argc)
-         {fprintf(stderr,"wwwoffle: The '-c' argument requires a configuration file name.\n"); exit(1);}
-
-       if(host)
-         {fprintf(stderr,"wwwoffle: The '-p' and '-c' options cannot be used together.\n"); exit(1);}
-
-       config_file=argv[i];
-
-       argv[i-1]=NULL;
-       argv[i]=NULL;
-       continue;
-      }
-
     if(!strcmp(argv[i],"-on") || !strcmp(argv[i],"-online"))
       {
        if(action!=None)
@@ -319,6 +205,24 @@ int main(int argc, char** argv)
        continue;
       }
 
+    if(!strcmp(argv[i],"-dump"))
+      {
+       if(action!=None)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Dump;
+       argv[i]=NULL;
+       continue;
+      }
+
+    if(!strcmp(argv[i],"-cyclelog"))
+      {
+       if(action!=None)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=CycleLog;
+       argv[i]=NULL;
+       continue;
+      }
+
     if(!strcmp(argv[i],"-purge"))
       {
        if(action!=None)
@@ -346,15 +250,171 @@ int main(int argc, char** argv)
        continue;
       }
 
+    /* Get options */
+
+    if(!strcmp(argv[i],"-F"))
+      {
+       if(action!=None && action!=Get)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Get;
+       if(force)
+         {fprintf(stderr,"wwwoffle: Only one '-F' argument may be given.\n"); exit(1);}
+       force=1;
+       argv[i]=NULL;
+       continue;
+      }
+
+    if(!strncmp(argv[i],"-g",(size_t)2))
+      {
+       if(action!=None && action!=Get)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Get;
+       if(strchr(argv[i]+2,'S'))
+         {
+          if(stylesheets)
+            {fprintf(stderr,"wwwoffle: Only one '-gS' argument may be given.\n"); exit(1);}
+          stylesheets=1;
+         }
+       else if(strchr(argv[i]+2,'i'))
+         {
+          if(images)
+            {fprintf(stderr,"wwwoffle: Only one '-gi' argument may be given.\n"); exit(1);}
+          images=1;
+         }
+       else if(strchr(argv[i]+2,'f'))
+         {
+          if(frames)
+            {fprintf(stderr,"wwwoffle: Only one '-gf' argument may be given.\n"); exit(1);}
+          frames=1;
+         }
+       else if(strchr(argv[i]+2,'s'))
+         {
+          if(scripts)
+            {fprintf(stderr,"wwwoffle: Only one '-gs' argument may be given.\n"); exit(1);}
+          scripts=1;
+         }
+       else if(strchr(argv[i]+2,'o'))
+         {
+          if(objects)
+            {fprintf(stderr,"wwwoffle: Only one '-go' argument may be given.\n"); exit(1);}
+          objects=1;
+         }
+       else if(argv[i][2]==0)
+         {
+          if(stylesheets || images || frames || scripts || objects)
+            {fprintf(stderr,"wwwoffle: Cannot have '-g' and '-g[Sifso]' together.\n"); exit(1);}
+          else if(nothing)
+            {fprintf(stderr,"wwwoffle: Only one '-g' argument can be given.\n"); exit(1);}
+          nothing=1;
+         }
+       else
+         {fprintf(stderr,"wwwoffle: The '-g' option does allow '%c' following it.\n",argv[i][2]); exit(1);}
+
+       argv[i]=NULL;
+       continue;
+      }
+
+    if(!strncmp(argv[i],"-R",(size_t)2) || !strncmp(argv[i],"-r",(size_t)2) || !strncmp(argv[i],"-d",(size_t)2))
+      {
+       if(action!=None && action!=Get)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Get;
+
+       if(depth)
+         {fprintf(stderr,"wwwoffle: Only one '-d', '-r' or '-R' argument may be given.\n"); exit(1);}
+
+       if(argv[i][1]=='d')
+          recursive_mode=1;
+       else if(argv[i][1]=='r')
+          recursive_mode=2;
+       else /* argv[i][1]=='R' */
+          recursive_mode=3;
+
+       if(argv[i][2])
+          depth=atoi(&argv[i][2]);
+       else if(i<(argc-1) && (atoi(argv[i+1]) || argv[i+1][0]=='0'))
+          depth=atoi(argv[i+1]);
+       else
+          depth=1;
+       if(depth<0)
+         {fprintf(stderr,"wwwoffle: The '-%c' argument may only be followed by non-negative integer.\n",argv[i][1]); exit(1);}
+
+       if(!argv[i][2] && i<(argc-1) && (atoi(argv[i+1]) || argv[i+1][0]=='0'))
+          argv[i++]=NULL;
+
+       argv[i]=NULL;
+       continue;
+      }
+
+    /* Server & port number / config file options */
+
+    if(!strcmp(argv[i],"-p"))
+      {
+       char *colon;
+
+       if(++i>=argc)
+         {fprintf(stderr,"wwwoffle: The '-p' argument requires a hostname and optionally a port number.\n"); exit(1);}
+
+       if(config_file)
+         {fprintf(stderr,"wwwoffle: The '-p' and '-c' options cannot be used together.\n"); exit(1);}
+
+       host=argv[i];
+
+       if(*host=='[')
+         {
+          char *square=strchr(host,']');
+          if(square)
+             colon=strchr(square,':');
+          else
+             colon=NULL;
+         }
+       else
+          colon=strchr(host,':');
+
+       if(colon)
+         {
+          *colon=0;
+          port=atoi(colon+1);
+
+          if(port<=0 || port>=65536)
+            {fprintf(stderr,"wwwoffle: The port number %d in -p option '%s' is invalid.\n",port,argv[i]); exit(1);}
+         }
+
+       argv[i-1]=NULL;
+       argv[i]=NULL;
+       continue;
+      }
+
+    if(!strcmp(argv[i],"-c"))
+      {
+       if(++i>=argc)
+         {fprintf(stderr,"wwwoffle: The '-c' argument requires a configuration file name.\n"); exit(1);}
+
+       if(host)
+         {fprintf(stderr,"wwwoffle: The '-p' and '-c' options cannot be used together.\n"); exit(1);}
+
+       config_file=argv[i];
+
+       argv[i-1]=NULL;
+       argv[i]=NULL;
+       continue;
+      }
+
+    /* Unknown options */
+
     if(argv[i][0]=='-' && argv[i][1])
       {
        fprintf(stderr,"wwwoffle: Unknown option '%s'.\n\n",argv[i]);
        usage(0);
       }
 
+    /* URLs */
+
     add_url_file(argv[i]);
    }
 
+ /* Check special conditions depending on the mode */
+
  if(action==None)
     action=Get;
 
@@ -363,14 +423,24 @@ int main(int argc, char** argv)
     fprintf(stderr,"wwwoffle: The -post and -put options require exactly one URL.\n\n");
     usage(0);
    }
-
- if((action==Output || action==OutputWithHeader) && n_url_file_list!=1)
+ else if((action==Output || action==OutputWithHeader) && n_url_file_list!=1)
    {
     fprintf(stderr,"wwwoffle: The -o and -O options require exactly one URL.\n\n");
     usage(0);
    }
+ else if(action==Get && n_url_file_list==0)
+   {
+    fprintf(stderr,"wwwoffle: No URLs were specified to request.\n\n");
+    usage(0);
+   }
+ else if(action!=Get && action!=Output && action!=OutputWithHeader && n_url_file_list!=0)
+   {
+    fprintf(stderr,"wwwoffle: The -online, -autodial, -offline, -fetch, -config, -dump, -cyclelog,\n"
+                   "          -purge, -status and -kill options must have no other arguments.\n\n");
+    usage(0);
+   }
 
- /* Initialise things. */
+ /* Check the environment variable. */
 
  if(!config_file && !host && (env=getenv("WWWOFFLE_PROXY")))
    {
@@ -378,51 +448,72 @@ int main(int argc, char** argv)
        config_file=env;
     else
       {
-       char *colon;
-       char *hoststr,*portstr;
+       char *colon1,*colon2;
 
        host=(char*)malloc(strlen(env)+1);
        strcpy(host,env);
 
-       SplitHostPort(host,&hoststr,&portstr);
-       host=hoststr;
-
-       if(portstr)
+       if(*host=='[')
          {
-          if((colon=strchr(portstr,':')))
+          char *square=strchr(host,']');
+          if(square)
+             colon1=strchr(square,':');
+          else
+             colon1=NULL;
+         }
+       else
+          colon1=strchr(host,':');
+
+       if(colon1)
+         {
+          *colon1=0;
+          colon2=strchr(colon1+1,':');
+
+          if(colon2)
             {
-             *colon=0;
-             if(action==Get || action==Output || action==OutputWithHeader)
-                port=atoi(portstr);
+             *colon2=0;
+             if(action==Get || action==Post || action==Put || action==Output || action==OutputWithHeader)
+                port=atoi(colon1+1);
              else
-                port=atoi(colon+1);
+                port=atoi(colon2+1);
             }
           else
-             port=atoi(portstr);
+             port=atoi(colon1+1);
 
           if(port<=0 || port>=65536)
-            {fprintf(stderr,"wwwoffle: The port number %d '%s' is invalid.\n",port,env); exit(1);}
+            {fprintf(stderr,"wwwoffle: The port number %d in WWWOFFLE_PROXY variable '%s' is invalid.\n",port,env); exit(1);}
          }
       }
    }
 
+ /* Load the configuration file. */
+
  InitErrorHandler("wwwoffle",0,1);
 
  InitConfigurationFile(config_file);
 
  if(config_file)
    {
-    init_buffer(2);
+    init_io(STDERR_FILENO);
 
-    if(ReadConfigurationFile(2))
+    if(ReadConfigurationFile(STDERR_FILENO))
        PrintMessage(Fatal,"Error in configuration file '%s'.",config_file);
+
+    finish_io(STDERR_FILENO);
    }
 
- if(!host)
-    host=GetLocalHost(0);
+ /* Check the host and port specified. */
 
- wwwoffle_port=ConfigInteger(WWWOFFLE_Port);
- http_port=ConfigInteger(HTTP_Port);
+ if(!host)
+    host=GetLocalHost();
+
+ if(!port)
+   {
+    if(action==Get || action==Post || action==Put || action==Output || action==OutputWithHeader)
+       port=ConfigInteger(HTTP_Port);
+    else
+       port=ConfigInteger(WWWOFFLE_Port);
+   }
 
  /* The connections to the WWWOFFLE server. */
 
@@ -431,11 +522,12 @@ int main(int argc, char** argv)
     int socket;
     char *line=NULL;
 
-    socket=OpenClientSocket(host,port?port:wwwoffle_port);
-    init_buffer(socket);
+    socket=OpenClientSocket(host,port);
 
     if(socket==-1)
-       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:wwwoffle_port);
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port);
+
+    init_io(socket);
 
     /* Send the message. */
 
@@ -452,6 +544,10 @@ int main(int argc, char** argv)
        write_string(socket,"WWWOFFLE FETCH\r\n");
     else if(action==Config)
        write_string(socket,"WWWOFFLE CONFIG\r\n");
+    else if(action==Dump)
+       write_string(socket,"WWWOFFLE DUMP\r\n");
+    else if(action==CycleLog)
+       write_string(socket,"WWWOFFLE CYCLELOG\r\n");
     else if(action==Purge)
        write_string(socket,"WWWOFFLE PURGE\r\n");
     else if(action==Status)
@@ -466,22 +562,25 @@ int main(int argc, char** argv)
        fputs(line,stdout);
        fflush(stdout);
 
-       if(!strncmp("WWWOFFLE Incorrect Password",line,27))
+       if(!strncmp("WWWOFFLE Incorrect Password",line,(size_t)27))
           exitval=3;
 
-       if(action==Online && !strncmp("WWWOFFLE Already Online",line,23))
+       if(action==Online && !strncmp("WWWOFFLE Already Online",line,(size_t)23))
           exitval=1;
-       else if(action==Autodial && !strncmp("WWWOFFLE Already in Autodial Mode",line,33))
+       else if(action==Autodial && !strncmp("WWWOFFLE Already in Autodial Mode",line,(size_t)33))
           exitval=1;
-       else if(action==Offline && !strncmp("WWWOFFLE Already Offline",line,24))
+       else if(action==Offline && !strncmp("WWWOFFLE Already Offline",line,(size_t)24))
           exitval=1;
-       else if(action==Fetch && (!strncmp("WWWOFFLE Already Fetching",line,25) ||
-                            !strncmp("WWWOFFLE Must be online or autodial to fetch",line,44)))
+       else if(action==Fetch && (!strncmp("WWWOFFLE Already Fetching",line,(size_t)25) ||
+                                 !strncmp("WWWOFFLE Must be online or autodial to fetch",line,(size_t)44)))
           exitval=1;
-       else if(action==Config && !strncmp("Configuration file syntax error",line,31))
+       else if(action==Config && !strncmp("Configuration file syntax error",line,(size_t)31))
+          exitval=1;
+       else if(action==CycleLog && !strncmp("WWWOFFLE Has No Log File",line,(size_t)24))
           exitval=1;
       }
 
+    finish_io(socket);
     CloseSocket(socket);
    }
 
@@ -497,23 +596,24 @@ int main(int argc, char** argv)
        if(strcmp(url_file_list[i],"-") && stat(url_file_list[i],&buf))
          {
           int socket;
-          char buffer[READ_BUFFER_SIZE];
+          char buffer[IO_BUFFER_SIZE];
 
           Url=SplitURL(url_file_list[i]);
 
-          if(!Url->Protocol)
+          if(!IsProtocolHandled(Url))
             {
              PrintMessage(Warning,"Cannot request protocol '%s'.",Url->proto);
              continue;
             }
 
-          socket=OpenClientSocket(host,port?port:http_port);
-          init_buffer(socket);
+          socket=OpenClientSocket(host,port);
 
           if(socket==-1)
-             PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:http_port);
+             PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port);
 
-          if(recursive_mode || depth || force || stylesheets || images || frames || scripts || objects)
+          init_io(socket);
+
+          if(recursive_mode || depth || force || stylesheets || images || frames || scripts || objects || nothing)
             {
              char *limit="";
 
@@ -521,14 +621,14 @@ int main(int argc, char** argv)
                 limit="";
              else if(recursive_mode==2)
                {
-                limit=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+5);
-                sprintf(limit,"%s://%s/",Url->proto,Url->host);
+                limit=(char*)malloc(strlen(Url->proto)+strlen(Url->hostport)+5);
+                sprintf(limit,"%s://%s/",Url->proto,Url->hostport);
                }
              else if(recursive_mode==1)
                {
                 char *p;
-                limit=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(Url->path)+5);
-                sprintf(limit,"%s://%s%s",Url->proto,Url->host,Url->path);
+                limit=(char*)malloc(strlen(Url->proto)+strlen(Url->hostport)+strlen(Url->path)+5);
+                sprintf(limit,"%s://%s%s",Url->proto,Url->hostport,Url->path);
                 p=limit+strlen(limit)-1;
                 while(p>limit && *p!='/')
                    *p--=0;
@@ -536,17 +636,15 @@ int main(int argc, char** argv)
              else
                 depth=0;
 
-             refresh=CreateRefreshPath(Url,limit,depth,
-                                       force,
-                                       stylesheets,images,frames,scripts,objects);
+             refresh=CreateRefreshPath(Url,limit,depth,force,stylesheets,images,frames,iframes,scripts,objects);
 
              if(*limit)
                 free(limit);
 
-             printf("Getting: %s (with recursive options).\n",Url->name);
+             printf("Requesting: %s (with recursive options).\n",Url->name);
             }
           else
-             printf("Getting: %s\n",Url->name);
+             printf("Requesting: %s\n",Url->name);
 
           if(Url->user)
             {
@@ -558,17 +656,16 @@ int main(int argc, char** argv)
              if(refresh)
                 write_formatted(socket,"GET %s HTTP/1.0\r\n"
                                        "Authorization: Basic %s\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
                                 refresh,userpass2);
              else
                 write_formatted(socket,"GET /refresh/?%s HTTP/1.0\r\n"
-                                       "Authorization: Basic %s\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
-                                URLEncodeFormArgs(Url->name),userpass2);
+                                URLEncodeFormArgs(Url->file));
 
              free(userpass1);
              free(userpass2);
@@ -577,21 +674,22 @@ int main(int argc, char** argv)
             {
              if(refresh)
                 write_formatted(socket,"GET %s HTTP/1.0\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
                                 refresh);
              else
                 write_formatted(socket,"GET /refresh/?%s HTTP/1.0\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
                                 URLEncodeFormArgs(Url->name));
             }
 
-          while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
+          while(read_data(socket,buffer,IO_BUFFER_SIZE)>0)
              ;
 
+          finish_io(socket);
           CloseSocket(socket);
 
           free(refresh);
@@ -600,13 +698,11 @@ int main(int argc, char** argv)
        else if(!strcmp(url_file_list[i],"-") || S_ISREG(buf.st_mode))
          {
           int file;
-          char *buffer=(char*)malloc(strlen(url_file_list[i])+256);
-          char **links;
+          URL **links;
 
           if(strcmp(url_file_list[i],"-"))
             {
              file=open(url_file_list[i],O_RDONLY);
-             init_buffer(file);
 
              if(file==-1)
                {PrintMessage(Warning,"Cannot open file '%s' for reading.",url_file_list[i]);continue;}
@@ -615,28 +711,32 @@ int main(int argc, char** argv)
             }
           else
             {
-             file=fileno(stdin);
-             init_buffer(file);
+             file=0;
 
              printf("Reading: <stdin>\n");
             }
 
-          strcpy(buffer,"file://localhost");
+          init_io(file);
+
           if(*url_file_list[i]=='/')
-             strcat(buffer,url_file_list[i]);
+             Url=CreateURL("file","localhost",url_file_list[i],NULL,NULL,NULL);
           else
             {
+             char *fullpath=(char*)malloc(strlen(url_file_list[i])+PATH_MAX+2);
              char cwd[PATH_MAX+1];
 
-             if(!getcwd(cwd,PATH_MAX))
-             cwd[0]=0;
+             if(!getcwd(cwd,(size_t)PATH_MAX))
+                cwd[0]=0;
 
-             strcat(buffer,cwd);
-             strcat(buffer,"/");
-             strcat(buffer,url_file_list[i]);
+             strcpy(fullpath,cwd);
+             strcat(fullpath,"/");
+             strcat(fullpath,url_file_list[i]);
+
+             Url=CreateURL("file","localhost",fullpath,NULL,NULL,NULL);
+
+             free(fullpath);
             }
 
-          Url=SplitURL(buffer);
           ParseHTML(file,Url);
 
           if(stylesheets && (links=GetReferences(RefStyleSheet)))
@@ -648,6 +748,9 @@ int main(int argc, char** argv)
           if(frames && (links=GetReferences(RefFrame)))
              add_url_list(links);
 
+          if(iframes && (links=GetReferences(RefIFrame)))
+             add_url_list(links);
+
           if(scripts && (links=GetReferences(RefScript)))
              add_url_list(links);
 
@@ -661,7 +764,7 @@ int main(int argc, char** argv)
              add_url_list(links);
 
           FreeURL(Url);
-          free(buffer);
+          finish_io(file);
           if(file!=0)
              close(file);
          }
@@ -672,29 +775,32 @@ int main(int argc, char** argv)
    {
     URL *Url;
     int socket,n,length=0;
-    char buffer[READ_BUFFER_SIZE];
-    char *data=(char*)malloc(READ_BUFFER_SIZE+1);
+    char buffer[IO_BUFFER_SIZE];
+    char *data=(char*)malloc((size_t)(IO_BUFFER_SIZE+1));
 
     Url=SplitURL(url_file_list[0]);
 
-    if(!Url->Protocol)
+    if(!IsProtocolHandled(Url))
        PrintMessage(Fatal,"Cannot post or put protocol '%s'.",Url->proto);
 
-    socket=OpenClientSocket(host,port?port:http_port);
-    init_buffer(socket);
+    socket=OpenClientSocket(host,port);
 
     if(socket==-1)
-       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:http_port);
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port);
+
+    init_io(socket);
 
     if(action==Post)
        printf("Posting: %s\n",Url->name);
     else
        printf("Putting: %s\n",Url->name);
 
-    while((n=read_data(0,data+length,READ_BUFFER_SIZE))>0)
+    init_io(0);
+
+    while((n=read_data(0,data+length,IO_BUFFER_SIZE))>0)
       {
        length+=n;
-       data=(char*)realloc((void*)data,length+READ_BUFFER_SIZE+1);
+       data=(char*)realloc((void*)data,(size_t)(length+IO_BUFFER_SIZE+1));
       }
 
     if(action==Post)
@@ -716,7 +822,7 @@ int main(int argc, char** argv)
                                  "Authorization: Basic %s\r\n"
                                  "Content-Type: application/x-www-form-urlencoded\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,userpass2,length);
@@ -724,7 +830,7 @@ int main(int argc, char** argv)
           write_formatted(socket,"PUT %s HTTP/1.0\r\n"
                                  "Authorization: Basic %s\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,userpass2,length);
@@ -738,14 +844,14 @@ int main(int argc, char** argv)
           write_formatted(socket,"POST %s HTTP/1.0\r\n"
                                  "Content-Type: application/x-www-form-urlencoded\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,length);
        else
           write_formatted(socket,"PUT %s HTTP/1.0\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,length);
@@ -755,9 +861,10 @@ int main(int argc, char** argv)
 
     write_data(socket,"\r\n",2);
 
-    while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
+    while(read_data(socket,buffer,IO_BUFFER_SIZE)>0)
        ;
 
+    finish_io(socket);
     CloseSocket(socket);
 
     FreeURL(Url);
@@ -767,25 +874,24 @@ int main(int argc, char** argv)
    {
     URL *Url;
     int socket;
-    char *line=NULL,buffer[READ_BUFFER_SIZE];
+    char *line=NULL;
+    char buffer[IO_BUFFER_SIZE];
     int nbytes;
 
-    if(!n_url_file_list)
-       PrintMessage(Fatal,"No URL specified to output.");
-
     Url=SplitURL(url_file_list[0]);
 
-    if(!Url->Protocol)
+    if(!IsProtocolHandled(Url))
        PrintMessage(Fatal,"Cannot fetch data from protocol '%s'.",Url->proto);
 
-    socket=OpenClientSocket(host,port?port:http_port);
-    init_buffer(socket);
+    socket=OpenClientSocket(host,port);
 
     if(socket==-1)
-       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:http_port);
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port);
+
+    init_io(socket);
 
     if(action!=Output && action!=OutputWithHeader)
-       fprintf(stderr,"Getting: %s\n",Url->name);
+       fprintf(stderr,"Requesting: %s\n",Url->name);
 
     if(Url->user)
       {
@@ -796,7 +902,7 @@ int main(int argc, char** argv)
 
        write_formatted(socket,"GET %s HTTP/1.0\r\n"
                               "Authorization: Basic %s\r\n"
-                              "Pragma: wwwoffle\r\n"
+                              "Pragma: wwwoffle-client\r\n"
                               "Accept: */*\r\n"
                               "\r\n",
                        Url->name,userpass2);
@@ -806,7 +912,7 @@ int main(int argc, char** argv)
       }
     else
        write_formatted(socket,"GET %s HTTP/1.0\r\n"
-                              "Pragma: wwwoffle\r\n"
+                              "Pragma: wwwoffle-client\r\n"
                               "Accept: */*\r\n"
                               "\r\n",
                        Url->name);
@@ -846,13 +952,14 @@ int main(int argc, char** argv)
                 break;
             }
 
-          while((nbytes=read_data(socket,buffer,READ_BUFFER_SIZE))>0)
+          while((nbytes=read_data(socket,buffer,IO_BUFFER_SIZE))>0)
              fwrite(buffer,1,nbytes,stdout);
          }
       }
     else
        PrintMessage(Fatal,"Cannot read from wwwoffle server.");
 
+    finish_io(socket);
     CloseSocket(socket);
 
     FreeURL(Url);
@@ -894,7 +1001,7 @@ static void usage(int verbose)
  fprintf(stderr,
          "Usage: wwwoffle -h | --help | --version\n"
          "       wwwoffle -online | -autodial | -offline | -fetch\n"
-         "       wwwoffle -config | -purge | -status | -kill\n"
+         "       wwwoffle -config | -dump | -cyclelog | -purge | -status | -kill\n"
          "       wwwoffle [-o|-O] <url>\n"
          "       wwwoffle [-post|-put] <url>\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] <url> ...\n"
@@ -924,6 +1031,10 @@ static void usage(int verbose)
             "\n"
             "wwwoffle -config     : Force the server to re-read the configuration file.\n"
             "\n"
+            "wwwoffle -dump       : Force the server to dump the current configuration.\n"
+            "\n"
+            "wwwoffle -cyclelog   : Force the server to close and re-open the log file.\n"
+            "\n"
             "wwwoffle -status     : Query the server about its current status.\n"
             "\n"
             "wwwoffle -purge      : Force the server to purge pages from the cache.\n"
@@ -937,8 +1048,9 @@ static void usage(int verbose)
             " -o                  : Fetch the URL and output it on the standard output.\n"
             " -O                  : As above but include the HTTP header.\n"
             "\n"
-            " -g[Sisfo]           : Fetch the items included in the specified URLs.\n"
+            " -g[Sisfo]           : Fetch the items included in the specified URLs:\n"
             "                       (S=stylesheets, i=images, f=frames, s=scripts, o=objects)\n"
+            "                       When used without any options fetches none of these.\n"
             " -F                  : Force the url to be refreshed even if already cached.\n"
             " -(d|r|R)[<depth>]   : Fetch pages linked to the URLs and their links,\n"
             "                       going no more than <depth> steps (default 1).\n"
@@ -946,6 +1058,10 @@ static void usage(int verbose)
             "                        (-r => URLs on the same host)\n"
             "                        (-R => URLs on any host)\n"
             "\n"
+            "                      (If the -F, -(d|r|R) or -g[Sisfo] options are set they\n"
+            "                       override the options in the FetchOptions section of the\n"
+            "                       config file and only the -g[Sisfo] options are fetched.)\n"
+            "\n"
             "wwwoffle -post <url> : Create a request using the POST method, the data is read\n"
             "                       from stdin and appended to the request.  The user should\n"
             "                       ensure that the data is correctly url-encoded.\n"
@@ -976,22 +1092,16 @@ static void usage(int verbose)
 /*++++++++++++++++++++++++++++++++++++++
   Add a list of URLs to the list.
 
-  char **links The list of URLs to add.
+  URL **links The list of URLs to add.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void add_url_list(char **links)
+static void add_url_list(URL **links)
 {
  int i;
 
  for(i=0;links[i];i++)
-   {
-    URL *linkUrl=SplitURL(links[i]);
-
-    if(strcmp(linkUrl->proto,"file"))
-       add_url_file(linkUrl->name);
-
-    FreeURL(linkUrl);
-   }
+    if(strcmp(links[i]->proto,"file"))
+       add_url_file(links[i]->name);
 }
 
 
