@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/wwwoffled.c 2.58 2002/10/20 10:07:57 amb Exp $
+  $Header: /home/amb/CVS/wwwoffle/src/wwwoffled.c,v 2.83 2009-03-13 19:30:39 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7g.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.9f.
   A demon program to maintain the database and spawn the servers.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1996-2009 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -41,12 +41,14 @@
 #include <signal.h>
 #include <grp.h>
 
-#include "version.h"
 #include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
+#include "errors.h"
 #include "config.h"
 #include "sockets.h"
-#include "errors.h"
+#include "certificates.h"
+#include "version.h"
 
 
 static void usage(int verbose);
@@ -59,6 +61,7 @@ static void sighup(int signum);
 
 /*+ The server sockets that we listen on +*/
 int http_fd[2]={-1,-1},          /*+ for the HTTP connections. +*/
+    https_fd[2]={-1,-1},         /*+ for the HTTPS connections. +*/
     wwwoffle_fd[2]={-1,-1};      /*+ for the WWWOFFLE connections. +*/
 
 /*+ The online / offline /autodial status. +*/
@@ -90,14 +93,11 @@ int purge_pid=0;
 /*+ The current status, fetching or not. +*/
 int fetching=0;
 
-/*+ The file descriptor of the spool directory. +*/
-int fSpoolDir=-1;
+/*+ Set when the demon is to shutdown due to a sigexit. +*/
+static sig_atomic_t got_sigexit=0;
 
-/*+ Set to 1 when the demon is to shutdown. +*/
-int closedown=0;
-
-/*+ Set to 1 when the demon is to re-read config file due to a sighup. +*/
-static int readconfig=0;
+/*+ Set when the demon is to re-read config file due to a sighup. +*/
+static sig_atomic_t got_sighup=0;
 
 /*+ Remember that we got a SIGCHLD +*/
 static sig_atomic_t got_sigchld=0;
@@ -108,12 +108,15 @@ static int detached=1;
 /*+ True if the pid of the daemon should be printed at startup +*/
 static int print_pid=0;
 
+/*+ The name of the configuration file specified. +*/
+static char *config_file=NULL;
+
+/*+ The name of the log file specified. +*/
+char *log_file=NULL;
+
 /*+ True if the -f option was passed on the command line. +*/
 int nofork=0;
 
-/* Global variables used to pass on information about remote client to CGI. */
-extern char *client_hostname,*client_ip;
-
 
 /*++++++++++++++++++++++++++++++++++++++
   The main program.
@@ -124,7 +127,6 @@ int main(int argc, char** argv)
  int i;
  int err;
  struct stat buf;
- char *config_file=NULL;
  int uid,gid;
 
  /* Parse the command line options */
@@ -140,12 +142,26 @@ int main(int argc, char** argv)
     if(!strcmp(argv[i],"-d"))
       {
        detached=0;
-       if(i<(argc-1) && isdigit(argv[i+1][0]) && !argv[i+1][1])
+
+       if(i<(argc-1) && isdigit(argv[i+1][0]))
          {
-          DebuggingLevel=Fatal+1-atoi(argv[++i]);
-          if(DebuggingLevel<0)
-             DebuggingLevel=0;
+          StderrLevel=Fatal+1-atoi(argv[++i]);
+
+          if(StderrLevel<0 || StderrLevel>Fatal)
+            {fprintf(stderr,"wwwoffled: The '-d' option requires a number between 0 and %d.\n",Fatal+1); exit(1);}
          }
+
+       continue;
+      }
+
+    if(!strcmp(argv[i],"-l"))
+      {
+       if(++i>=argc)
+         {fprintf(stderr,"wwwoffled: The '-l' option requires a filename.\n"); exit(1);}
+       if(argv[i][0]!='/')
+         {fprintf(stderr,"wwwoffled: The '-l' option requires an absolute pathname.\n"); exit(1);}
+
+       log_file=argv[i];
        continue;
       }
 
@@ -174,6 +190,19 @@ int main(int argc, char** argv)
     fprintf(stderr,"wwwoffled: Unknown option '%s'.\n",argv[i]); exit(1);
    }
 
+ /* Combination options */
+
+ if(log_file)
+   {
+    if(nofork)
+       log_file=NULL;           /* -f overrides -l */
+    else
+       detached=1;              /* -l overrides -d */
+
+    if(StderrLevel==-1)
+       StderrLevel=Inform;      /* -l sets log level if no -d */
+   }
+
  /* Initialise things. */
 
  for(i=0;i<MAX_FETCH_SERVERS;i++)
@@ -182,33 +211,45 @@ int main(int argc, char** argv)
  for(i=0;i<MAX_SERVERS;i++)
     server_pids[i]=0;
 
+ if(log_file)
+    OpenErrorLog(log_file);
+
  InitErrorHandler("wwwoffled",0,1); /* use stderr and not syslog to start with. */
 
  /* Read the configuration file. */
 
  InitConfigurationFile(config_file);
 
- init_buffer(2);
- if(ReadConfigurationFile(2))
+ init_io(STDERR_FILENO);
+
+ if(ReadConfigurationFile(STDERR_FILENO))
     PrintMessage(Fatal,"Error in configuration file '%s'.",ConfigurationFileName());
 
+ finish_io(STDERR_FILENO);
+
  InitErrorHandler("wwwoffled",ConfigInteger(UseSyslog),1); /* enable syslog if requested. */
 
  /* Print a startup message. */
 
-#if USE_ZLIB
-#define ZLIB_STRING "with zlib"
-#else
-#define ZLIB_STRING "without zlib"
-#endif
-
 #if USE_IPV6
 #define IPV6_STRING "with ipv6"
 #else
 #define IPV6_STRING "without ipv6"
 #endif
 
- PrintMessage(Important,"WWWOFFLE Demon Version %s (%s,%s) started.",WWWOFFLE_VERSION,ZLIB_STRING,IPV6_STRING);
+#if USE_ZLIB
+#define ZLIB_STRING "with zlib"
+#else
+#define ZLIB_STRING "without zlib"
+#endif
+
+#if USE_GNUTLS
+#define GNUTLS_STRING "with gnutls"
+#else
+#define GNUTLS_STRING "without gnutls"
+#endif
+
+ PrintMessage(Important,"WWWOFFLE Demon Version %s (%s,%s,%s) started.",WWWOFFLE_VERSION,IPV6_STRING,ZLIB_STRING,GNUTLS_STRING);
  PrintMessage(Inform,"WWWOFFLE Read Configuration File '%s'.",ConfigurationFileName());
 
  /* Change the user and group. */
@@ -219,29 +260,32 @@ int main(int argc, char** argv)
  if(uid!=-1)
     seteuid(0);
 
+ if(log_file && (uid!=-1 || gid!=-1))
+    chown(log_file,(uid_t)uid,(gid_t)gid);
+
  if(gid!=-1)
    {
 #if HAVE_SETGROUPS
     if(getuid()==0 || geteuid()==0)
-       if(setgroups(0, NULL)<0)
+       if(setgroups(0,NULL)<0)
           PrintMessage(Fatal,"Cannot clear supplementary group list [%!s].");
 #endif
 
 #if HAVE_SETRESGID
-    if(setresgid(gid,gid,gid)<0)
+    if(setresgid((gid_t)gid,(gid_t)gid,(gid_t)gid)<0)
        PrintMessage(Fatal,"Cannot set real/effective/saved group id to %d [%!s].",gid);
 #else
     if(geteuid()==0)
       {
-       if(setgid(gid)<0)
+       if(setgid((gid_t)gid)<0)
           PrintMessage(Fatal,"Cannot set group id to %d [%!s].",gid);
       }
     else
       {
 #if HAVE_SETREGID
-       if(setregid(getegid(),gid)<0)
+       if(setregid(getegid(),(gid_t)gid)<0)
           PrintMessage(Fatal,"Cannot set effective group id to %d [%!s].",gid);
-       if(setregid(gid,-1)<0)
+       if(setregid((gid_t)gid,(gid_t)~1)<0)
           PrintMessage(Fatal,"Cannot set real group id to %d [%!s].",gid);
 #else
        PrintMessage(Fatal,"Must be root to totally change group id.");
@@ -253,20 +297,20 @@ int main(int argc, char** argv)
  if(uid!=-1)
    {
 #if HAVE_SETRESUID
-    if(setresuid(uid,uid,uid)<0)
+    if(setresuid((uid_t)uid,(uid_t)uid,(uid_t)uid)<0)
        PrintMessage(Fatal,"Cannot set real/effective/saved user id to %d [%!s].",uid);
 #else
     if(geteuid()==0)
       {
-       if(setuid(uid)<0)
+       if(setuid((uid_t)uid)<0)
           PrintMessage(Fatal,"Cannot set user id to %d [%!s].",uid);
       }
     else
       {
 #if HAVE_SETREUID
-       if(setreuid(geteuid(),uid)<0)
+       if(setreuid(geteuid(),(uid_t)uid)<0)
           PrintMessage(Fatal,"Cannot set effective user id to %d [%!s].",uid);
-       if(setreuid(uid,-1)<0)
+       if(setreuid((uid_t)uid,(uid_t)~1)<0)
           PrintMessage(Fatal,"Cannot set real user id to %d [%!s].",uid);
 #else
        PrintMessage(Fatal,"Must be root to totally change user id.");
@@ -281,6 +325,25 @@ int main(int argc, char** argv)
  if(geteuid()==0 || getegid()==0)
     PrintMessage(Warning,"Running with root user or group privileges is not recommended.");
 
+ /* Create, Change to and open the spool directory. */
+
+ umask(0);
+
+ if(stat(ConfigString(SpoolDir),&buf))
+   {
+    err=mkdir(ConfigString(SpoolDir),(mode_t)ConfigInteger(DirPerm));
+    if(err==-1 && errno!=EEXIST)
+       PrintMessage(Fatal,"Cannot create spool directory %s [%!s].",ConfigString(SpoolDir));
+    stat(ConfigString(SpoolDir),&buf);
+   }
+
+ if(!S_ISDIR(buf.st_mode))
+    PrintMessage(Fatal,"The spool directory %s is not a directory.",SpoolDir);
+
+ err=ChangeToSpoolDir(ConfigString(SpoolDir));
+ if(err==-1)
+    PrintMessage(Fatal,"Cannot change to spool directory %s [%!s].",ConfigString(SpoolDir));
+
  /* Bind the HTTP proxy socket(s). */
 
 #if USE_IPV6
@@ -299,7 +362,7 @@ int main(int argc, char** argv)
       {
 #if USE_IPV6
        if(http_fd[1]!=-1 && ConfigString(Bind_IPv4) && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0") &&
-                            ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"0:0:0:0:0:0:0:0"))
+                            ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"[0:0:0:0:0:0:0:0]"))
           PrintMessage(Warning,"Cannot create HTTP IPv4 server socket (but the IPv6 one might accept IPv4 connections).");
        else
           PrintMessage(Fatal,"Cannot create HTTP IPv4 server socket.");
@@ -314,10 +377,61 @@ int main(int argc, char** argv)
 #if USE_IPV6
     PrintMessage(Fatal,"The IPv4 and IPv6 HTTP sockets were not bound; are they disabled in the config file?");
 #else
-    PrintMessage(Fatal,"The HTTP socket was not bound; are they disabled in the config file?");
+    PrintMessage(Fatal,"The HTTP socket was not bound; is it disabled in the config file?");
 #endif
    }
 
+ /* Bind the HTTPS socket(s). */
+
+#if USE_GNUTLS
+
+ if(LoadRootCredentials())
+   {
+    PrintMessage(Warning,"Failed to read (or create if needed) the WWWOFFLE root certificates.");
+   }
+
+ if(LoadTrustedCertificates())
+   {
+    PrintMessage(Warning,"Failed to read in any trusted certificates.");
+   }
+
+#if USE_IPV6
+ if(ConfigString(Bind_IPv6))
+   {
+    https_fd[1]=OpenServerSocket(ConfigString(Bind_IPv6),ConfigInteger(HTTPS_Port));
+    if(https_fd[1]==-1)
+       PrintMessage(Fatal,"Cannot create HTTPS IPv6 server socket.");
+   }
+#endif
+
+ if(ConfigString(Bind_IPv4))
+   {
+    https_fd[0]=OpenServerSocket(ConfigString(Bind_IPv4),ConfigInteger(HTTPS_Port));
+    if(https_fd[0]==-1)
+      {
+#if USE_IPV6
+       if(https_fd[1]!=-1 && ConfigString(Bind_IPv4) && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0") &&
+                             ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"[0:0:0:0:0:0:0:0]"))
+          PrintMessage(Warning,"Cannot create HTTPS IPv4 server socket (but the IPv6 one might accept IPv4 connections).");
+       else
+          PrintMessage(Fatal,"Cannot create HTTPS IPv4 server socket.");
+#else
+       PrintMessage(Fatal,"Cannot create HTTPS server socket.");
+#endif
+      }
+   }
+
+ if(https_fd[0]==-1 && https_fd[1]==-1)
+   {
+#if USE_IPV6
+    PrintMessage(Fatal,"The IPv4 and IPv6 HTTPS sockets were not bound; are they disabled in the config file?");
+#else
+    PrintMessage(Fatal,"The HTTPS socket was not bound; is it disabled in the config file?");
+#endif
+   }
+
+#endif /* USE_GNUTLS */
+
  /* Bind the WWWOFFLE control socket(s). */
 
 #if USE_IPV6
@@ -336,7 +450,7 @@ int main(int argc, char** argv)
       {
 #if USE_IPV6
        if(wwwoffle_fd[1]!=-1 && ConfigString(Bind_IPv4) && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0") &&
-                                ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"0:0:0:0:0:0:0:0"))
+                                ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"[0:0:0:0:0:0:0:0]"))
           PrintMessage(Warning,"Cannot create WWWOFFLE IPv4 server socket (but the IPv6 one might accept IPv4 connections).");
        else
           PrintMessage(Fatal,"Cannot create WWWOFFLE IPv4 server socket.");
@@ -351,44 +465,30 @@ int main(int argc, char** argv)
 #if USE_IPV6
     PrintMessage(Fatal,"The IPv4 and IPv6 WWWOFFLE sockets were not bound; are they disabled in the config file?");
 #else
-    PrintMessage(Fatal,"The WWWOFFLE socket was not bound; are they disabled in the config file?");
+    PrintMessage(Fatal,"The WWWOFFLE socket was not bound; is it disabled in the config file?");
 #endif
    }
 
- /* Create, Change to and open the spool directory. */
-
- umask(0);
-
- if(stat(ConfigString(SpoolDir),&buf))
-   {
-    err=mkdir(ConfigString(SpoolDir),(mode_t)ConfigInteger(DirPerm));
-    if(err==-1 && errno!=EEXIST)
-       PrintMessage(Fatal,"Cannot create spool directory %s [%!s].",ConfigString(SpoolDir));
-    stat(ConfigString(SpoolDir),&buf);
-   }
-
- if(!S_ISDIR(buf.st_mode))
-    PrintMessage(Fatal,"The spool directory %s is not a directory.",SpoolDir);
-
- err=chdir(ConfigString(SpoolDir));
- if(err==-1)
-    PrintMessage(Fatal,"Cannot change to spool directory %s [%!s].",ConfigString(SpoolDir));
-
-#if !defined(__CYGWIN__)
- fSpoolDir=open(".",O_RDONLY);
- if(fSpoolDir==-1)
-    PrintMessage(Fatal,"Cannot open the spool directory '%s' [%!s].",ConfigString(SpoolDir));
-#endif
-
  /* Detach from terminal */
 
  if(detached)
    {
     demoninit();
+
     PrintMessage(Important,"Detached from terminal and changed pid to %d.",getpid());
-    InitErrorHandler("wwwoffled",1,0); /* pid changes after detaching, enable syslog, disable stderr. */
+
+    if(log_file)
+       InitErrorHandler("wwwoffled",-1,-1); /* pid changes after detaching, keep stderr as was. */
+    else
+       InitErrorHandler("wwwoffled",-1, 0); /* pid changes after detaching, disable stderr. */
+
+    if(!log_file)
+       close(STDERR_FILENO);
    }
 
+ close(STDIN_FILENO);
+ close(STDOUT_FILENO);
+
  install_sighandlers();
 
  max_servers=ConfigInteger(MaxServers);
@@ -396,6 +496,8 @@ int main(int argc, char** argv)
 
  /* Loop around waiting for connections. */
 
+ PrintMessage(Inform,"WWWOFFLE Ready to accept connections.");
+
  do
    {
     struct timeval tv;
@@ -410,12 +512,18 @@ int main(int argc, char** argv)
 #endif
           if(http_fd[nfd]>=nfds)
              nfds=http_fd[nfd]+1;
+          if(https_fd[nfd]>=nfds)
+             nfds=https_fd[nfd]+1;
           if(wwwoffle_fd[nfd]>=nfds)
              nfds=wwwoffle_fd[nfd]+1;
 
           if(n_servers<max_servers)
+            {
              if(http_fd[nfd]!=-1)
                 FD_SET(http_fd[nfd],&readfd);
+             if(https_fd[nfd]!=-1)
+                FD_SET(https_fd[nfd],&readfd);
+            }
 
           if(wwwoffle_fd[nfd]!=-1)
              FD_SET(wwwoffle_fd[nfd],&readfd);
@@ -438,12 +546,17 @@ int main(int argc, char** argv)
              int port,client;
 
              client=AcceptConnect(wwwoffle_fd[nfd]);
-             init_buffer(client);
 
              if(client>=0)
                {
+                init_io(client);
+                configure_io_timeout(client,ConfigInteger(SocketTimeout),ConfigInteger(SocketTimeout));
+
                 if(SocketRemoteName(client,&host,&ip,&port))
+                  {
+                   finish_io(client);
                    CloseSocket(client);
+                  }
                 else
                   {
                    char *canonical_ip=CanonicaliseHost(ip);
@@ -455,16 +568,19 @@ int main(int argc, char** argv)
                       CommandConnect(client);
 
                       if(fetch_fd!=client)
+                        {
+                         finish_io(client);
                          CloseSocket(client);
+                        }
                      }
                    else
                      {
                       PrintMessage(Warning,"WWWOFFLE Connection rejected from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+                      finish_io(client);
                       CloseSocket(client);
                      }
 
-                   if(canonical_ip!=ip)
-                      free(canonical_ip);
+                   free(canonical_ip);
                   }
                }
             }
@@ -475,10 +591,12 @@ int main(int argc, char** argv)
              int port,client;
 
              client=AcceptConnect(http_fd[nfd]);
-             init_buffer(client);
 
              if(client>=0)
                {
+                init_io(client);
+                configure_io_timeout(client,ConfigInteger(SocketTimeout),ConfigInteger(SocketTimeout));
+
                 if(!SocketRemoteName(client,&host,&ip,&port))
                   {
                    char *canonical_ip=CanonicaliseHost(ip);
@@ -486,37 +604,83 @@ int main(int argc, char** argv)
                    if(IsAllowedConnectHost(host) || IsAllowedConnectHost(canonical_ip))
                      {
                       PrintMessage(Inform,"HTTP Proxy connection from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
-
-                      client_hostname=host;
-                      client_ip=canonical_ip;
-
-                      ForkServer(client,1);
+                      ForkServer(client);
                      }
                    else
                       PrintMessage(Warning,"HTTP Proxy connection rejected from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
 
-                   if(canonical_ip!=ip)
-                      free(canonical_ip);
+                   free(canonical_ip);
                   }
 
                 if(nofork)
-                   closedown=1;
+                   got_sigexit=1;
 
                 if(!nofork)
+                  {
+                   finish_io(client);
                    CloseSocket(client);
+                  }
                }
             }
+
+#if USE_GNUTLS
+          if(https_fd[nfd]!=-1 && FD_ISSET(https_fd[nfd],&readfd))
+            {
+             char *host,*ip;
+             int port,client;
+
+             client=AcceptConnect(https_fd[nfd]);
+
+             if(client>=0)
+               {
+                init_io(client);
+                configure_io_timeout(client,ConfigInteger(SocketTimeout),ConfigInteger(SocketTimeout));
+
+                if(!SocketRemoteName(client,&host,&ip,&port))
+                  {
+                   char *canonical_ip=CanonicaliseHost(ip);
+
+                   if(IsAllowedConnectHost(host) || IsAllowedConnectHost(canonical_ip))
+                     {
+                      PrintMessage(Inform,"HTTPS Proxy connection from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+                      ForkServer(client);
+                     }
+                   else
+                      PrintMessage(Warning,"HTTPS Proxy connection rejected from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+
+                   free(canonical_ip);
+                  }
+
+                if(nofork)
+                   got_sigexit=1;
+
+                if(!nofork)
+                  {
+                   finish_io(client);
+                   CloseSocket(client);
+                  }
+               }
+            }
+#endif /* USE_GNUTLS */
+
 #if USE_IPV6
          }
 #endif
       }
 
-    if(readconfig)
+    if(got_sighup)
       {
-       readconfig=0;
+       got_sighup=0;
 
        PrintMessage(Important,"SIGHUP signalled.");
 
+       if(log_file)
+         {
+          PrintMessage(Important,"Closing and opening log file.");
+
+          OpenErrorLog(log_file);
+         }
+
        PrintMessage(Important,"WWWOFFLE Re-reading Configuration File.");
 
        if(ReadConfigurationFile(-1))
@@ -531,6 +695,7 @@ int main(int argc, char** argv)
        int isserver=0;
 
        /* To avoid race conditions, reset the flag before fetching the status */
+
        got_sigchld=0;
 
        while((pid=waitpid(-1,&status,WNOHANG))>0)
@@ -570,6 +735,7 @@ int main(int argc, char** argv)
                }
 
           /* Check if the child that terminated is one of the fetching wwwoffles */
+
           for(i=0;i<max_fetch_servers;i++)
              if(fetch_pids[i]==pid)
                {
@@ -581,10 +747,10 @@ int main(int argc, char** argv)
           if(exitval==3)
              fetching=0;
 
-          if(exitval==4 && online==1)
+          if(exitval==4 && online!=0)
              fetching=1;
 
-          if(online!=1)
+          if(online==0)
              fetching=0;
          }
 
@@ -596,30 +762,32 @@ int main(int argc, char** argv)
        start fetch servers to look for jobs in the spool directory. */
 
     while(fetching && n_fetch_servers<max_fetch_servers && n_servers<max_servers)
-       ForkServer(fetch_fd,0);
+       ForkServer(fetch_fd);
 
     if(fetch_fd!=-1 && !fetching && n_fetch_servers==0)
       {
        write_string(fetch_fd,"WWWOFFLE No more to fetch.\n");
+
+       finish_io(fetch_fd);
        CloseSocket(fetch_fd);
        fetch_fd=-1;
 
        PrintMessage(Important,"WWWOFFLE Fetch finished.");
 
-       ForkRunModeScript(ConfigString(RunFetch),"fetch","stop",fetch_fd);
+       ForkRunModeScript(ConfigString(RunFetch),"fetch","stop",-1);
       }
    }
- while(!closedown);
+ while(!got_sigexit);
 
  /* Close down and exit. */
 
- if(http_fd[0]!=-1) CloseSocket(http_fd[0]);
- if(wwwoffle_fd[0]!=-1) CloseSocket(wwwoffle_fd[0]);
+ /* These four sockets don't need finish_io() calling because they never
+    had init_io() called, they are just bound to a port listening. */
 
-#if USE_IPV6
+ if(http_fd[0]!=-1) CloseSocket(http_fd[0]);
  if(http_fd[1]!=-1) CloseSocket(http_fd[1]);
+ if(wwwoffle_fd[0]!=-1) CloseSocket(wwwoffle_fd[0]);
  if(wwwoffle_fd[1]!=-1) CloseSocket(wwwoffle_fd[1]);
-#endif
 
  if(!nofork)
    {
@@ -672,8 +840,8 @@ int main(int argc, char** argv)
 
  FinishConfigurationFile();
 
-#if !defined(__CYGWIN__)
- close(fSpoolDir);
+#if USE_GNUTLS
+ FreeLoadedCredentials();
 #endif
 
  return(0);
@@ -703,31 +871,34 @@ static void usage(int verbose)
 
  if(verbose)
     fprintf(stderr,
-            "(c) Andrew M. Bishop 1996,97,98,99,2000,01 [       amb@gedanken.demon.co.uk ]\n"
-            "                                           [http://www.gedanken.demon.co.uk/]\n"
+            "(c) Andrew M. Bishop 1996,97,98,99,2000,01,02,03 [       amb@gedanken.demon.co.uk ]\n"
+            "                                                 [http://www.gedanken.demon.co.uk/]\n"
             "\n");
 
  fprintf(stderr,
          "Usage: wwwoffled -h | --help | --version\n"
-         "       wwwoffled [-c <config-file>] [-d [<log-level>]]\n"
+         "       wwwoffled [-c <config-file>] [-d [<log-level>]] [-l <log-file>] [-f] [-p]\n"
          "\n");
 
  if(verbose)
     fprintf(stderr,
-            "   -h | --help         : Display this help.\n"
+            "   -h | --help      : Display this help.\n"
             "\n"
-            "   --version           : Display the version number of wwwoffled.\n"
+            "   --version        : Display the version number of wwwoffled.\n"
             "\n"
-            "   -c <config-file>    : The name of the configuration file to use.\n"
-            "                         (Default %s).\n"
+            "   -c <config-file> : The name of the configuration file to use.\n"
+            "                      (Default %s).\n"
             "\n"
-            "   -d [<log-level>]    : Do not detach from the terminal and use stderr.\n"
-            "                         0 <= log-level <= %d (default in config file).\n"
+            "   -d [<log-level>] : Do not detach from the terminal and use stderr.\n"
+            "                      0 <= log-level <= %d (default in config file).\n"
             "\n"
-            "   -f                  : Do not fork children for each HTTP request.\n"
-            "                         (Exits after first request, implies -d)\n"
+            "   -l <log-file>    : Detach from the terminal and use specified log file.\n"
+            "                      (Using log-level = 4 unless -d specifies different.)\n"
             "\n"
-            "   -p                  : Print the pid of wwwoffled on stdout (unless -d).\n"
+            "   -f               : Do not fork children for each HTTP request.\n"
+            "                      (Exits after first request, implies -d, overrides -l.)\n"
+            "\n"
+            "   -p               : Print the pid of wwwoffled on stdout (unless -d).\n"
             "\n",ConfigurationFileName(),Fatal+1);
 
  exit(0);
@@ -761,11 +932,6 @@ static void demoninit(void)
 
     exit(0);
    }
-
- /* Close fds */
- close(STDIN_FILENO);
- close(STDOUT_FILENO);
- close(STDERR_FILENO);
 }
 
 
@@ -779,7 +945,7 @@ static void install_sighandlers(void)
 
  /* SIGCHLD */
  action.sa_handler = sigchild;
- sigemptyset (&action.sa_mask);
+ sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  if(sigaction(SIGCHLD, &action, NULL) != 0)
     PrintMessage(Warning, "Cannot install SIGCHLD handler.");
@@ -807,7 +973,7 @@ static void install_sighandlers(void)
 
  /* SIGPIPE */
  action.sa_handler = SIG_IGN;
- sigemptyset (&action.sa_mask);
+ sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  if(sigaction(SIGPIPE, &action, NULL) != 0)
     PrintMessage(Warning, "Cannot ignore SIGPIPE.");
@@ -834,7 +1000,7 @@ static void sigchild(/*@unused@*/ int signum)
 
 static void sigexit(/*@unused@*/ int signum)
 {
- closedown=1;
+ got_sigexit=1;
 }
 
 
@@ -846,5 +1012,5 @@ static void sigexit(/*@unused@*/ int signum)
 
 static void sighup(/*@unused@*/ int signum)
 {
- readconfig=1;
+ got_sighup=1;
 }
